import pytest

from postgis import MultiPoint

pytestmark = pytest.mark.asyncio


@pytest.mark.parametrize('expected', [
    ((30, 10), (10, 30)),
])
async def test_multipoint_should_round(connection, expected):
    geom = MultiPoint(expected, srid=4326)
    await connection.execute('INSERT INTO multipoint_async (geom) VALUES ($1)',
                             geom)
    geom = await connection.fetchval('SELECT geom FROM multipoint_async WHERE '
                                     'geom=$1', geom, column=0)
    assert geom.coords == expected



from postgis import Point, LineString, Polygon, GeometryCollection

POLYGON = Polygon((
    ((1, 2), (3, 4), (5, 6), (1, 2)),
    ((2, 3), (4, 5), (6, 7), (2, 3))
))

COLLECTION = [
    Point(1, 2),
    LineString(((1, 2), (3, 4))),
    POLYGON
]


def test_geometrycollection_should_round(cursor):
    params = [GeometryCollection(COLLECTION, srid=4326)]
    cursor.execute('INSERT INTO geometrycollection (geom) VALUES (%s)', params)
    cursor.execute('SELECT geom FROM geometrycollection WHERE geom=%s', params)
    geom = cursor.fetchone()[0]
    assert geom == COLLECTION



import pytest

from postgis import MultiLineString


@pytest.mark.parametrize('expected', [
    (((30, 10), (10, 30)), ((40, 10), (10, 40))),
])
def test_multilinestring_should_round(cursor, expected):
    params = [MultiLineString(expected, srid=4326)]
    cursor.execute('INSERT INTO multilinestring (geom) VALUES (%s)', params)
    cursor.execute('SELECT geom FROM multilinestring WHERE geom=%s', params)
    geom = cursor.fetchone()[0]
    assert geom.coords == expected



import pytest

from postgis import MultiPolygon

MULTI = (
            (
                ((35, 10), (45, 45), (15, 40), (10, 20), (35, 10)),
                ((20, 30), (35, 35), (30, 20), (20, 30))
            ),
            (
                ((36, 10), (46, 45), (16, 40), (16, 20), (36, 10)),
                ((21, 30), (36, 35), (36, 20), (21, 30))
            ),
        )


@pytest.mark.parametrize('expected', [
    MULTI,
])
def test_multipolygon_should_round(cursor, expected):
    params = [MultiPolygon(expected, srid=4326)]
    cursor.execute('INSERT INTO multipolygon (geom) VALUES (%s)', params)
    cursor.execute('SELECT geom FROM multipolygon WHERE geom=%s', params)
    geom = cursor.fetchone()[0]
    assert geom.coords == expected



from .geometry import Geometry


class Multi(Geometry):

    __slots__ = ['geoms', 'srid']
    SUBCLASS = None

    def __init__(self, geoms, srid=None):
        self.geoms = [self.SUBCLASS(g, srid=srid) for g in geoms]
        if srid:
            self.srid = srid

    def __iter__(self):
        return iter(self.geoms)

    @property
    def has_z(self):
        return self[0].has_z

    @property
    def has_m(self):
        return self[0].has_m

    def __getitem__(self, item):
        return self.geoms[item]

    @classmethod
    def from_ewkb_body(cls, reader, srid=None):
        return cls([reader.read() for index in range(reader.read_int())], srid)

    @property
    def wkt_coords(self):
        return ', '.join('({})'.format(g.wkt_coords) for g in self)

    def write_ewkb_body(self, writer):
        writer.write_int(len(self.geoms))
        for geom in self:
            geom.write_ewkb(writer)

    @property
    def coords(self):
        return tuple(g.coords for g in self)



try:
    # Do not make psycopg2 a requirement.
    from psycopg2.extensions import ISQLQuote
except ImportError:
    print('psycopg2 not installed')


from .ewkb import Reader, Typed, Writer
from .geojson import GeoJSON


class Geometry(object, metaclass=Typed):

    @property
    def has_z(self):
        return False

    @property
    def has_m(self):
        return False

    @property
    def has_srid(self):
        return hasattr(self, 'srid')

    @staticmethod
    def from_ewkb(value, cursor=None):
        if not value:
            return None
        return Reader.from_hex(value)

    def to_ewkb(self):
        return Writer.to_hex(self).decode()

    def write_ewkb(self, writer):
        self.write_ewkb_body(writer.clone(self))

    def text(self):
        return "ST_GeometryFromText('{}', {})".format(self.wkt, self.srid)

    # Psycopg2 interface.
    def __conform__(self, protocol):
        if protocol is ISQLQuote:
            return self

    def getquoted(self):
        return "'{}'".format(self.to_ewkb())
    # End Psycopg2 interface.

    def __str__(self):
        return self.wkt

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.wkt)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            other = other.coords
        return self.coords == other

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def wkt(self):
        return "{}({})".format(self.name.upper(), self.wkt_coords)

    @property
    def geojson(self):
        return GeoJSON({
            'type': self.name,
            'coordinates': self.coords
        })



from .linestring import LineString
from .multi import Multi


class Polygon(Multi):

    TYPE = 3
    SUBCLASS = LineString

    @classmethod
    def from_ewkb_body(cls, reader, srid=None):
        return cls([LineString.from_ewkb_body(reader)
                   for index in range(reader.read_int())], srid)

    def write_ewkb_body(self, writer):
        writer.write_int(len(self.geoms))
        for geom in self:
            geom.write_ewkb_body(writer)



import pytest

from postgis import Polygon


@pytest.mark.parametrize('expected', [
    (((35, 10), (45, 45), (15, 40), (10, 20), (35, 10)), ((20, 30), (35, 35), (30, 20), (20, 30))),  # noqa
])
def test_polyggon_should_round(cursor, expected):
    params = [Polygon(expected, srid=4326)]
    cursor.execute('INSERT INTO polygon (geom) VALUES (%s)', params)
    cursor.execute('SELECT geom FROM polygon WHERE geom=%s', params)
    geom = cursor.fetchone()[0]
    assert geom.coords == expected



import pytest

from postgis import Polygon

pytestmark = pytest.mark.asyncio


@pytest.mark.parametrize('expected', [
    (((35, 10), (45, 45), (15, 40), (10, 20), (35, 10)), ((20, 30), (35, 35), (30, 20), (20, 30))),  # noqa
])
async def test_polygon_should_round(connection, expected):
    geom = Polygon(expected, srid=4326)
    await connection.execute('INSERT INTO polygon_async (geom) VALUES ($1)',
                             geom)
    geom = await connection.fetchval('SELECT geom FROM polygon_async WHERE '
                                     'geom=$1', geom, column=0)
    assert geom.coords == expected



import psycopg2

import pytest
from postgis import (GeometryCollection, LineString, MultiLineString,
                     MultiPoint, MultiPolygon, Point, Polygon, register)

db = psycopg2.connect(dbname="test")
cur = db.cursor()
register(db)

geoms = [Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon,
         GeometryCollection]


def pytest_configure(config):
    tpl = 'CREATE TABLE IF NOT EXISTS {} ("geom" geometry({}) NOT NULL)'
    for geom in geoms:
        name = geom.__name__
        cur.execute(tpl.format(name.lower(), name))


def pytest_unconfigure(config):
    for geom in geoms:
        cur.execute('DROP TABLE {}'.format(geom.__name__.lower()))
    db.commit()
    db.close()


@pytest.fixture
def cursor():
    # Make sure tables are clean.
    for geom in geoms:
        cur.execute('TRUNCATE TABLE {}'.format(geom.__name__.lower()))
    return cur



"Postgis helpers for psycopg2 and asyncpg."
from .geometry import Geometry
from .geometrycollection import GeometryCollection
from .linestring import LineString
from .multilinestring import MultiLineString
from .multipoint import MultiPoint
from .multipolygon import MultiPolygon
from .point import Point
from .polygon import Polygon
try:
    from .psycopg import register  # Retrocompat.
except ImportError:
    pass

__all__ = ['Geometry', 'register', 'Point', 'LineString', 'Polygon',
           'MultiPoint', 'MultiLineString', 'MultiPolygon',
           'GeometryCollection']

try:
    import pkg_resources
except ImportError:  # pragma: no cover
    pass
else:
    if __package__:
        VERSION = pkg_resources.get_distribution(__package__).version



import pytest

from postgis import Point

pytestmark = pytest.mark.asyncio


@pytest.mark.parametrize('expected', [
    (1, -2),
    (-1.123456789, 2.987654321),
])
async def test_point_should_round(connection, expected):
    point = Point(*expected, srid=4326)
    await connection.execute('INSERT INTO point_async (geom) VALUES ($1)',
                             point)
    geom = await connection.fetchval('SELECT geom FROM point_async WHERE '
                                     'geom=$1', point, column=0)
    assert geom.coords == expected


async def test_point_with_geography_column(connection):
    await connection.execute('CREATE TABLE geo ("geom" geography(PointZ))')
    point = Point(1, 2, 3, srid=4326)
    await connection.execute('INSERT INTO geo (geom) VALUES ($1)', point)
    geom = await connection.fetchval('SELECT geom FROM geo WHERE geom=$1',
                                     point, column=0)
    assert geom.coords == (1, 2, 3)
    await connection.execute('DROP TABLE geo')



import asyncpg
import pytest
from postgis import (GeometryCollection, LineString, MultiLineString,
                     MultiPoint, MultiPolygon, Point, Polygon)
from postgis.asyncpg import register

geoms = [Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon,
         GeometryCollection]


@pytest.yield_fixture
async def connection():
    conn = await asyncpg.connect('postgresql://postgres@localhost/test')
    await register(conn)
    tpl = 'CREATE TABLE IF NOT EXISTS {}_async ("geom" geometry({}) NOT NULL)'
    for geom in geoms:
        name = geom.__name__
        await conn.execute(tpl.format(name.lower(), name))
    yield conn
    await conn.close()



from .geometry import Geometry


async def register(connection):

    def encoder(value):
        if not isinstance(value, Geometry):
            raise ValueError('Geometry value must subclass Geometry class')
        return value.to_ewkb()

    def decoder(value):
        return Geometry.from_ewkb(value)

    await connection.set_type_codec(
        'geography', encoder=encoder, decoder=decoder)
    await connection.set_type_codec(
        'geometry', encoder=encoder, decoder=decoder)



import pytest

from postgis import MultiPoint


@pytest.mark.parametrize('expected', [
    ((30, 10), (10, 30)),
])
def test_multipoint_should_round(cursor, expected):
    params = [MultiPoint(expected, srid=4326)]
    cursor.execute('INSERT INTO multipoint (geom) VALUES (%s)', params)
    cursor.execute('SELECT geom FROM multipoint WHERE geom=%s', params)
    geom = cursor.fetchone()[0]
    assert geom.coords == expected



from psycopg2 import extensions

from .geometry import Geometry


def register(connection):
    if isinstance(connection, extensions.cursor):
        # Retrocompat.
        cursor = connection
    else:
        cursor = connection.cursor()
    cursor.execute("SELECT NULL::geometry")
    oid = cursor.description[0][1]
    GEOMETRY = extensions.new_type((oid, ), "GEOMETRY", Geometry.from_ewkb)
    extensions.register_type(GEOMETRY)

    cursor.execute("SELECT NULL::geography")
    oid = cursor.description[0][1]
    GEOGRAPHY = extensions.new_type((oid, ), "GEOGRAPHY", Geometry.from_ewkb)
    extensions.register_type(GEOGRAPHY)



import pytest

from postgis import Point


@pytest.mark.parametrize('expected', [
    (1, -2),
    (-1.123456789, 2.987654321),
])
def test_point_should_round(cursor, expected):
    params = [Point(*expected, srid=4326)]
    cursor.execute('INSERT INTO point (geom) VALUES (%s)', params)
    cursor.execute('SELECT geom FROM point WHERE geom=%s', params)
    geom = cursor.fetchone()[0]
    assert geom.coords == expected


@pytest.mark.parametrize('expected', [
    (1, -2, 3),
    (-1.123456789, 2.987654321, 231),
    (1, -2, 0),
])
def test_point_geography_column_should_round(cursor, expected):
    cursor.execute('CREATE TABLE geography_point ("geom" geography(PointZ))')
    params = [Point(*expected, srid=4326)]
    cursor.execute('INSERT INTO geography_point (geom) VALUES (%s)', params)
    cursor.execute('SELECT geom FROM geography_point WHERE geom=%s', params)
    geom = cursor.fetchone()[0]
    assert geom.coords == expected
    cursor.execute('DROP TABLE geography_point')



from .point import Point
from .multi import Multi


class LineString(Multi):

    TYPE = 2
    SUBCLASS = Point

    @classmethod
    def from_ewkb_body(cls, reader, srid=None):
        return cls([Point.from_ewkb_body(reader)
                   for index in range(reader.read_int())], srid)

    def write_ewkb_body(self, writer):
        writer.write_int(len(self.geoms))
        for geom in self:
            geom.write_ewkb_body(writer)



from .geometry import Geometry


class Point(Geometry):

    __slots__ = ['x', 'y', 'z', 'm', 'srid']
    TYPE = 1

    def __init__(self, x, y=None, z=None, m=None, srid=None):
        if y is None and isinstance(x, (tuple, list)):
            x, y, *extra = x
            if extra:
                z, *extra = extra
                if extra:
                    m = extra[0]
        self.x = float(x)
        self.y = float(y)
        self.z = int(z) if z is not None else None
        self.m = int(m) if m is not None else None
        if srid is not None:
            self.srid = srid

    def __getitem__(self, item):
        if item in (0, 'x'):
            return self.x
        elif item in (1, 'y'):
            return self.y
        elif item in (2, 'z'):
            return self.z
        elif item in (3, 'm'):
            return self.m

    def __iter__(self):
        return iter(self.values())

    _keys = ['x', 'y', 'z', 'm']

    def keys(self):
        return [k for k in self._keys if self[k] is not None]

    def values(self):
        return tuple(self[k] for k in self.keys())

    @property
    def has_z(self):
        return self.z is not None

    @classmethod
    def from_ewkb_body(cls, reader, srid=None):
        return cls(reader.read_double(), reader.read_double(),
                   reader.read_double() if reader.has_z else None,
                   reader.read_double() if reader.has_m else None,
                   srid)

    def write_ewkb_body(self, writer):
        writer.write_double(self.x)
        writer.write_double(self.y)
        if self.z is not None:
            writer.write_double(self.z)
        if self.m is not None:
            writer.write_double(self.m)

    @property
    def wkt_coords(self):
        return ' '.join(map(str, self.coords))

    @property
    def coords(self):
        return self.values()



import pytest

from postgis import LineString


@pytest.mark.parametrize('expected', [
    ((30, 10), (10, 30), (40, 40)),
])
def test_linestring_should_round(cursor, expected):
    params = [LineString(expected, srid=4326)]
    cursor.execute('INSERT INTO linestring (geom) VALUES (%s)', params)
    cursor.execute('SELECT geom FROM linestring WHERE geom=%s', params)
    geom = cursor.fetchone()[0]
    assert geom.coords == expected



from .geometry import Geometry
from .geojson import GeoJSON


class GeometryCollection(Geometry):

    TYPE = 7

    def __init__(self, geoms, srid=None):
        for geom in geoms:
            if not isinstance(geom, Geometry):
                raise ValueError('{} is not instance of Geometry'.format(geom))
        self.geoms = list(geoms)
        if srid:
            self.srid = srid

    def __iter__(self):
        return self.geoms.__iter__()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            other = other.geoms
        return self.geoms == other

    @property
    def has_z(self):
        return self[0].has_z

    @property
    def has_m(self):
        return self[0].has_m

    def __getitem__(self, item):
        return self.geoms[item]

    @classmethod
    def from_ewkb_body(cls, reader, srid=None):
        return cls([reader.read() for index in range(reader.read_int())], srid)

    def write_ewkb_body(self, writer):
        writer.write_int(len(self.geoms))
        for geom in self:
            geom.write_ewkb(writer)

    @property
    def wkt_coords(self):
        return ', '.join(g.wkt for g in self)

    @property
    def geojson(self):
        return GeoJSON({
            'type': self.name,
            'geometries': [g.geojson for g in self]
        })



import pytest

from postgis import Point, LineString, Polygon, GeometryCollection

pytestmark = pytest.mark.asyncio


POLYGON = Polygon((
    ((1, 2), (3, 4), (5, 6), (1, 2)),
    ((2, 3), (4, 5), (6, 7), (2, 3))
))

COLLECTION = [
    Point(1, 2),
    LineString(((1, 2), (3, 4))),
    POLYGON
]


async def test_geometrycollection_should_round(connection):
    geom = GeometryCollection(COLLECTION, srid=4326)
    await connection.execute('INSERT INTO geometrycollection_async (geom) '
                             'VALUES ($1)', geom)
    geom = await connection.fetchval('SELECT geom '
                                     'FROM geometrycollection_async '
                                     'WHERE geom=$1', geom, column=0)
    assert geom == COLLECTION



import pytest

from postgis import MultiLineString

pytestmark = pytest.mark.asyncio


@pytest.mark.parametrize('expected', [
    (((30, 10), (10, 30)), ((40, 10), (10, 40))),
])
async def test_multilinestring_should_round(connection, expected):
    geom = MultiLineString(expected, srid=4326)
    await connection.execute('INSERT INTO multilinestring_async (geom) '
                             'VALUES ($1)', geom)
    geom = await connection.fetchval('SELECT geom FROM multilinestring_async '
                                     'WHERE geom=$1', geom, column=0)
    assert geom.coords == expected



"""Pyscopg and asyncpg helpers to work with PostGIS."""

import glob
from setuptools import setup, find_packages, Extension
from codecs import open  # To use a consistent encoding
from os import path
import sys

HERE = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


def is_pkg(line):
    return line and not line.startswith(('--', 'git', '#'))


def list_modules(dirname):
    filenames = glob.glob(path.join(dirname, '*.py'))

    module_names = []
    for name in filenames:
        module, ext = path.splitext(path.basename(name))
        if module != '__init__':
            module_names.append(module)

    return module_names


try:
    from Cython.Distutils import build_ext
    CYTHON = True
except ImportError:
    sys.stdout.write('\nNOTE: Cython not installed. python-postgis will '
                     'still work fine, but may run a bit slower.\n\n')
    CYTHON = False
    cmdclass = {}
    ext_modules = []
else:
    ext_modules = [
        Extension('postgis.' + ext, [path.join('postgis', ext + '.py')])
        for ext in list_modules(path.join(HERE, 'postgis'))]

    cmdclass = {'build_ext': build_ext}


VERSION = (1, 0, 3)

setup(
    name='postgis',
    version=".".join(map(str, VERSION)),
    description=__doc__,
    long_description=long_description,
    url="https://github.com/yohanboniface/python-postgis",
    author='Yohan Boniface',
    author_email='yohan.boniface@data.gouv.fr',
    license='WTFPL',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: GIS',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='psycopg postgis gis asyncpg',
    packages=find_packages(exclude=['tests']),
    extras_require={'test': ['pytest'], 'docs': 'mkdocs'},
    include_package_data=True,
    cmdclass=cmdclass,
    ext_modules=ext_modules,
)



import pytest

from postgis import MultiPolygon

pytestmark = pytest.mark.asyncio

MULTI = (
            (
                ((35, 10), (45, 45), (15, 40), (10, 20), (35, 10)),
                ((20, 30), (35, 35), (30, 20), (20, 30))
            ),
            (
                ((36, 10), (46, 45), (16, 40), (16, 20), (36, 10)),
                ((21, 30), (36, 35), (36, 20), (21, 30))
            ),
        )


@pytest.mark.parametrize('expected', [
    MULTI,
])
async def test_multipolygon_should_round(connection, expected):
    geom = MultiPolygon(expected, srid=4326)
    await connection.execute('INSERT INTO multipolygon_async (geom) '
                             'VALUES ($1)', geom)
    geom = await connection.fetchval('SELECT geom FROM multipolygon_async '
                                     'WHERE geom=$1', geom, column=0)
    assert geom.coords == expected



import binascii
from io import BytesIO
import struct


class Typed(type):

    types = {}

    def __new__(mcs, name, bases, attrs, **kwargs):
        cls = super().__new__(mcs, name, bases, attrs, **kwargs)
        if hasattr(cls, 'TYPE'):
            Typed.types[cls.TYPE] = cls
        return cls

    def __call__(cls, *args, **kwargs):
        # Allow to pass an instance as first argument, for blind casting.
        if args and isinstance(args[0], cls):
            return args[0]
        return super().__call__(*args, **kwargs)


class Reader:

    __slots__ = ['stream', 'endianness', 'has_z', 'has_m']

    def __init__(self, stream):
        self.stream = stream

    def clone(self):
        return type(self)(self.stream)

    def read(self):
        # https://en.wikipedia.org/wiki/Well-known_text#Well-known_binary
        byte_order = self.stream.read(1)
        if byte_order == b'\x00':
            self.endianness = b'>'
        elif byte_order == b'\x01':
            self.endianness = b'<'
        else:
            raise Exception('invalid encoding')

        type_ = self.read_int()
        self.has_z = bool(type_ & 0x80000000)
        self.has_m = bool(type_ & 0x40000000)
        srid = self.read_int() if bool(type_ & 0x20000000) else None
        type_ &= 0x1fffffff

        try:
            class_ = Typed.types[type_]
        except KeyError:
            raise ValueError('unsupported geometry type {0}'.format(type_))
        else:
            return class_.from_ewkb_body(self, srid)

    def read_int(self):
        return struct.unpack(self.endianness + b'I', self.stream.read(4))[0]

    def read_double(self):
        return struct.unpack(self.endianness + b'd', self.stream.read(8))[0]

    @classmethod
    def from_hex(cls, value):
        return cls(BytesIO(binascii.a2b_hex(value))).read()


class Writer:

    __slots__ = ['stream']

    def __init__(self, geometry, stream=None):
        self.stream = stream or BytesIO()
        try:
            type_ = geometry.TYPE
        except AttributeError:
            raise ValueError('Unknown geometry {}'.format(geometry.__class__))

        # Little endian.
        self.stream.write(b'\x01')
        self.write_int(
            type_ |
            (0x80000000 if geometry.has_z else 0) |
            (0x40000000 if geometry.has_m else 0) |
            (0x20000000 if geometry.has_srid else 0))
        if geometry.has_srid:
            self.write_int(geometry.srid)

    def write_int(self, value):
        self.stream.write(struct.pack(b'<I', value))

    def write_double(self, value):
        self.stream.write(struct.pack(b'<d', value))

    def clone(self, geometry):
        return type(self)(geometry, self.stream)

    @classmethod
    def to_hex(cls, value):
        writer = cls(value)
        value.write_ewkb_body(writer)
        return binascii.b2a_hex(writer.stream.getvalue()).upper()
